﻿/** Example:
select arc.build_partitional('MY_SCHEMA','MY_TABLE','TIME_FIELD','DAY_TO_PARTITION', COUNT_OF_PARTITION);
select arc.build_partitional('schema1','table1','dt','2011-6-7', 5);
*/

create or replace function build_partitional(in text, in text, in text, in date, in int, out boolean)
  returns boolean as
$$
  my ($schema,$table,$field,$date1,$count) = @_;
  (my $name = $date1) =~ s/-/_/g;
 
  # Убедимся что, что таблица ${schema}.${table}_${name} не существует
  my $sql = "select * from pg_tables where schemaname='${schema}' and tablename='${table}_${name}'";
  my $q = spi_exec_query($sql);
  if ( defined($q->{rows}[0]) ) {
    return  false;
  }

  my $rv = spi_exec_query("select date('$date1'::date+interval'1 day') as a", 1);
  my $date2 = $rv->{rows}[0]->{a}; # следующий после date1 день

  ## Найдём первичный ключ для родительской таблицы, чтобы создать аналогичные ключ для дочерней таблицы
  $sql = "select k.constraint_name, k.column_name
	from information_schema.key_column_usage k
	join information_schema.table_constraints c on k.constraint_schema = c.constraint_schema and k.constraint_name = c.constraint_name
	where c.constraint_type = 'PRIMARY KEY' and 
	k.constraint_schema='${schema}' and k.table_name = '${table}'
	order by k.ordinal_position";
  $q = spi_query($sql);
  my $pkey;
  while (my $row = spi_fetchrow($q)) {    
    $pkey .= ",$row->{'column_name'}";
  }
  $pkey =~ s/,//; # удалим одну лишнюю запятую вначале
  $pkey = "constraint ${schema}_${table}_${name}_pkey primary key ($pkey)";
  
  ## Найдём все внешние ключи для родительской таблицы, чтобы создать аналогичные ключи для дочерней таблицы.
  ## Замечание: расчитано только на простые индексы (в будущем можно учесть и сложные индексы)
  $sql = "select k.table_name as k_table_name, k.column_name as k_column_name, t.table_name as f_table_name, l.column_name as f_column_name, t.constraint_name, r.delete_rule, r.update_rule
	from information_schema.key_column_usage k
	join information_schema.table_constraints       c on k.constraint_name = c.constraint_name
	join information_schema.constraint_table_usage  t on k.constraint_name = t.constraint_name
	join information_schema.constraint_column_usage l on k.constraint_name = l.constraint_name
	join information_schema.referential_constraints r on k.constraint_name = r.constraint_name
	where c.constraint_type = 'FOREIGN KEY' and 
	k.constraint_schema='${schema}' and k.table_name = '${table}'";
  $q = spi_query($sql);
  my $fkey;
  while (my $row = spi_fetchrow($q)) {    
    $fkey .= ",constraint ${schema}_${table}_${name}_$row->{'constraint_name'}";
    $fkey .= " foreign key ($row->{'k_column_name'}) references ${schema}.$row->{'f_table_name'} ($row->{'f_column_name'})";
    $fkey .= " on delete $row->{'delete_rule'} on update $row->{'update_rule'}";    
  }
  
  ## Для таблицы ${schema}.${table} создадим дочернюю таблицу ${schema}.${table}_${name}  
  $sql = "create table ${schema}.${table}_${name} ( 
      check(date '$date1' <= $field and $field < date '$date2'),
      $pkey
      $fkey
    ) inherits (${schema}.${table}) without oids";
  spi_exec_query($sql);
  
  ## Создадим кластер по этому индексу  
  $sql = "cluster ${schema}_${table}_${name}_pkey on ${schema}.${table}_${name}"; # primary key - это индекс, а значит можно сформировать по нему кластер
  spi_exec_query($sql);
  

  ## Создадим/поменяем триггерную функцию для таблицы ${schema}.${table}
  $sql = "select tablename from pg_tables where schemaname='${schema}' and tablename ~ '^${table}_.+' order by tablename desc limit $count";
  elog(NOTICE, "anselm: $sql"); # отладочное сообщение
  $q = spi_query($sql);
  $sql = "create or replace function ${schema}.${table}_before_insert() 
    returns trigger as 
    \$\$ 
    begin
    "; 
  
  while (defined (my $row = spi_fetchrow($q))) {
    $row->{'tablename'} =~ m/_(.+)/;
    my $suffix = $1;
    
    (my $d1 = $suffix) =~ s/_/-/g;
    my $rv = spi_exec_query("select date('$d1'::date+interval'1 day') as a", 1);
    my $d2 = $rv->{rows}[0]->{a}; # следующий после d1 день
    
    $sql .= "if date '$d1' <= new.$field and new.$field < date '$d2' then insert into ${schema}.${table}_$suffix values (new.*);
      else";
  }
  $sql .= " raise exception 'Value of \"$field\" out of range. Fix the ${schema}.${table}_before_insert() function!';
      end if;
      return null;
    end;
    \$\$
    language plpgsql";
  spi_exec_query($sql);

  ## Создадим (если надо) триггер для родительской таблицы
  $sql = "select * from pg_trigger where tgname = '${schema}_${table}_before_insert'";
  $q = spi_query($sql);
  unless (my $row = spi_fetchrow($q)) {
    $sql = "create trigger ${schema}_${table}_before_insert
	before insert
	on ${schema}.${table}
	for each row
	execute procedure ${schema}.${table}_before_insert()";
    spi_exec_query($sql);
  }
  
  ## Дадим те же права, что и на родительскую таблицу
  $sql = "select grantee,privilege_type from information_schema.table_privileges 
	where table_schema='${schema}' and table_name='${table}'";
  $q = spi_query($sql);
  $sql = "";
  while (my $row = spi_fetchrow($q)) {
    $sql .= "grant $row->{'privilege_type'} on table ${schema}.${table}_${name} to $row->{'grantee'};";
  }
  
  spi_exec_query($sql);

  
  return true;
$$
language plperl;

